package cloud

import (
	"encoding/json"
	"fmt"
	"strings"
	"testing"

	"github.com/stretchr/testify/assert"
)

type test struct {
	A string
}

func TestTransport_unmarshal_must_fail_with_old_format(t *testing.T) {
	msg := &test{
		A: "hello",
	}

	plainJSON, err := json.Marshal(msg)
	assert.NoError(t, err)

	assert.Error(t, unmarshal(plainJSON, &struct{}{}))
}

func TestTransport_small_message(t *testing.T) {
	msg := &test{
		A: "hello",
	}

	data, err := marshal(msg)
	assert.NoError(t, err)

	assert.True(t, len(data) < dataLimit)

	msg2 := &test{}
	assert.NoError(t, unmarshal(data, msg2))

	assert.Equal(t, msg, msg2)
}

func TestTransport_large_message(t *testing.T) {
	// concat example payloads until it exceeds the limit
	longPayload := strings.Repeat(examplePayload, dataLimit/len(examplePayload)+1)

	msg := &test{
		A: longPayload,
	}

	assert.True(t, len(longPayload) > dataLimit, "test setup failed, example data must be larger than limit")

	data, err := marshal(msg)
	assert.NoError(t, err)

	assert.True(t, len(data) < dataLimit, fmt.Sprintf("data length %d is larger than limit %d, compression expected", len(data), dataLimit))

	msg2 := &test{}
	assert.NoError(t, unmarshal(data, msg2))

	assert.Equal(t, msg, msg2)
}

const examplePayload = `{"action":"created","check_run":{"app":{"created_at":"2019-04-19T19:36:24Z","description":"","events":[],"external_url":"https://octocoders.github.io","html_url":"https://github.com/apps/octocoders-linter","id":29310,"name":"octocoders-linter","node_id":"MDM6QXBwMjkzMTA=","owner":{"avatar_url":"https://avatars1.githubusercontent.com/u/38302899?v=4","events_url":"https://api.github.com/users/Octocoders/events{/privacy}","followers_url":"https://api.github.com/users/Octocoders/followers","following_url":"https://api.github.com/users/Octocoders/following{/other_user}","gists_url":"https://api.github.com/users/Octocoders/gists{/gist_id}","gravatar_id":"","html_url":"https://github.com/Octocoders","id":38302899,"login":"Octocoders","node_id":"MDEyOk9yZ2FuaXphdGlvbjM4MzAyODk5","organizations_url":"https://api.github.com/users/Octocoders/orgs","received_events_url":"https://api.github.com/users/Octocoders/received_events","repos_url":"https://api.github.com/users/Octocoders/repos","site_admin":false,"starred_url":"https://api.github.com/users/Octocoders/starred{/owner}{/repo}","subscriptions_url":"https://api.github.com/users/Octocoders/subscriptions","type":"Organization","url":"https://api.github.com/users/Octocoders"},"permissions":{"administration":"write","checks":"write","contents":"write","deployments":"write","issues":"write","members":"write","metadata":"read","organization_administration":"write","organization_hooks":"write","organization_plan":"read","organization_projects":"write","organization_user_blocking":"write","pages":"write","pull_requests":"write","repository_hooks":"write","repository_projects":"write","statuses":"write","team_discussions":"write","vulnerability_alerts":"read"},"updated_at":"2019-04-19T19:36:56Z"},"check_suite":{"after":"ec26c3e57ca3a959ca5aad62de7213c562f8c821","app":{"created_at":"2019-04-19T19:36:24Z","description":"","events":[],"external_url":"https://octocoders.github.io","html_url":"https://github.com/apps/octocoders-linter","id":29310,"name":"octocoders-linter","node_id":"MDM6QXBwMjkzMTA=","owner":{"avatar_url":"https://avatars1.githubusercontent.com/u/38302899?v=4","events_url":"https://api.github.com/users/Octocoders/events{/privacy}","followers_url":"https://api.github.com/users/Octocoders/followers","following_url":"https://api.github.com/users/Octocoders/following{/other_user}","gists_url":"https://api.github.com/users/Octocoders/gists{/gist_id}","gravatar_id":"","html_url":"https://github.com/Octocoders","id":38302899,"login":"Octocoders","node_id":"MDEyOk9yZ2FuaXphdGlvbjM4MzAyODk5","organizations_url":"https://api.github.com/users/Octocoders/orgs","received_events_url":"https://api.github.com/users/Octocoders/received_events","repos_url":"https://api.github.com/users/Octocoders/repos","site_admin":false,"starred_url":"https://api.github.com/users/Octocoders/starred{/owner}{/repo}","subscriptions_url":"https://api.github.com/users/Octocoders/subscriptions","type":"Organization","url":"https://api.github.com/users/Octocoders"},"permissions":{"administration":"write","checks":"write","contents":"write","deployments":"write","issues":"write","members":"write","metadata":"read","organization_administration":"write","organization_hooks":"write","organization_plan":"read","organization_projects":"write","organization_user_blocking":"write","pages":"write","pull_requests":"write","repository_hooks":"write","repository_projects":"write","statuses":"write","team_discussions":"write","vulnerability_alerts":"read"},"updated_at":"2019-04-19T19:36:56Z"},"before":"6113728f27ae82c7b1a177c8d03f9e96e0adf246","conclusion":null,"created_at":"2019-05-15T15:20:31Z","head_branch":"changes","head_sha":"ec26c3e57ca3a959ca5aad62de7213c562f8c821","id":118578147,"node_id":"MDEwOkNoZWNrU3VpdGUxMTg1NzgxNDc=","pull_requests":[{"base":{"ref":"master","repo":{"id":186853002,"name":"Hello-World","url":"https://api.github.com/repos/Codertocat/Hello-World"},"sha":"f95f852bd8fca8fcc58a9a2d6c842781e32a215e"},"head":{"ref":"changes","repo":{"id":186853002,"name":"Hello-World","url":"https://api.github.com/repos/Codertocat/Hello-World"},"sha":"ec26c3e57ca3a959ca5aad62de7213c562f8c821"},"id":279147437,"number":2,"url":"https://api.github.com/repos/Codertocat/Hello-World/pulls/2"}],"status":"queued","updated_at":"2019-05-15T15:20:31Z","url":"https://api.github.com/repos/Codertocat/Hello-World/check-suites/118578147"},"completed_at":null,"conclusion":null,"deployment":{"created_at":"2021-02-18T08:22:48Z","description":null,"environment":"lab","id":326191728,"node_id":"MDEwOkRlcGxveW1lbnQzMjYxOTE3Mjg=","original_environment":"lab","repository_url":"https://api.github.com/repos/Codertocat/Hello-World","statuses_url":"https://api.github.com/repos/Codertocat/Hello-World/deployments/326191728/statuses","task":"deploy","updated_at":"2021-02-18T09:47:16Z","url":"https://api.github.com/repos/Codertocat/Hello-World/deployments/326191728"},"details_url":"https://octocoders.github.io","external_id":"","head_sha":"ec26c3e57ca3a959ca5aad62de7213c562f8c821","html_url":"https://github.com/Codertocat/Hello-World/runs/128620228","id":128620228,"name":"Octocoders-linter","node_id":"MDg6Q2hlY2tSdW4xMjg2MjAyMjg=","output":{"annotations_count":0,"annotations_url":"https://api.github.com/repos/Codertocat/Hello-World/check-runs/128620228/annotations","summary":null,"text":null,"title":null},"pull_requests":[{"base":{"ref":"master","repo":{"id":186853002,"name":"Hello-World","url":"https://api.github.com/repos/Codertocat/Hello-World"},"sha":"f95f852bd8fca8fcc58a9a2d6c842781e32a215e"},"head":{"ref":"changes","repo":{"id":186853002,"name":"Hello-World","url":"https://api.github.com/repos/Codertocat/Hello-World"},"sha":"ec26c3e57ca3a959ca5aad62de7213c562f8c821"},"id":279147437,"number":2,"url":"https://api.github.com/repos/Codertocat/Hello-World/pulls/2"}],"started_at":"2019-05-15T15:21:12Z","status":"queued","url":"https://api.github.com/repos/Codertocat/Hello-World/check-runs/128620228"},"repository":{"archive_url":"https://api.github.com/repos/Codertocat/Hello-World/{archive_format}{/ref}","archived":false,"assignees_url":"https://api.github.com/repos/Codertocat/Hello-World/assignees{/user}","blobs_url":"https://api.github.com/repos/Codertocat/Hello-World/git/blobs{/sha}","branches_url":"https://api.github.com/repos/Codertocat/Hello-World/branches{/branch}","clone_url":"https://github.com/Codertocat/Hello-World.git","collaborators_url":"https://api.github.com/repos/Codertocat/Hello-World/collaborators{/collaborator}","comments_url":"https://api.github.com/repos/Codertocat/Hello-World/comments{/number}","commits_url":"https://api.github.com/repos/Codertocat/Hello-World/commits{/sha}","compare_url":"https://api.github.com/repos/Codertocat/Hello-World/compare/{base}...{head}","contents_url":"https://api.github.com/repos/Codertocat/Hello-World/contents/{+path}","contributors_url":"https://api.github.com/repos/Codertocat/Hello-World/contributors","created_at":"2019-05-15T15:19:25Z","default_branch":"master","deployments_url":"https://api.github.com/repos/Codertocat/Hello-World/deployments","description":null,"disabled":false,"downloads_url":"https://api.github.com/repos/Codertocat/Hello-World/downloads","events_url":"https://api.github.com/repos/Codertocat/Hello-World/events","fork":false,"forks":1,"forks_count":1,"forks_url":"https://api.github.com/repos/Codertocat/Hello-World/forks","full_name":"Codertocat/Hello-World","git_commits_url":"https://api.github.com/repos/Codertocat/Hello-World/git/commits{/sha}","git_refs_url":"https://api.github.com/repos/Codertocat/Hello-World/git/refs{/sha}","git_tags_url":"https://api.github.com/repos/Codertocat/Hello-World/git/tags{/sha}","git_url":"git://github.com/Codertocat/Hello-World.git","has_downloads":true,"has_issues":true,"has_pages":true,"has_projects":true,"has_wiki":true,"homepage":null,"hooks_url":"https://api.github.com/repos/Codertocat/Hello-World/hooks","html_url":"https://github.com/Codertocat/Hello-World","id":186853002,"issue_comment_url":"https://api.github.com/repos/Codertocat/Hello-World/issues/comments{/number}","issue_events_url":"https://api.github.com/repos/Codertocat/Hello-World/issues/events{/number}","issues_url":"https://api.github.com/repos/Codertocat/Hello-World/issues{/number}","keys_url":"https://api.github.com/repos/Codertocat/Hello-World/keys{/key_id}","labels_url":"https://api.github.com/repos/Codertocat/Hello-World/labels{/name}","language":"Ruby","languages_url":"https://api.github.com/repos/Codertocat/Hello-World/languages","license":null,"merges_url":"https://api.github.com/repos/Codertocat/Hello-World/merges","milestones_url":"https://api.github.com/repos/Codertocat/Hello-World/milestones{/number}","mirror_url":null,"name":"Hello-World","node_id":"MDEwOlJlcG9zaXRvcnkxODY4NTMwMDI=","notifications_url":"https://api.github.com/repos/Codertocat/Hello-World/notifications{?since,all,participating}","open_issues":2,"open_issues_count":2,"owner":{"avatar_url":"https://avatars1.githubusercontent.com/u/21031067?v=4","events_url":"https://api.github.com/users/Codertocat/events{/privacy}","followers_url":"https://api.github.com/users/Codertocat/followers","following_url":"https://api.github.com/users/Codertocat/following{/other_user}","gists_url":"https://api.github.com/users/Codertocat/gists{/gist_id}","gravatar_id":"","html_url":"https://github.com/Codertocat","id":21031067,"login":"Codertocat","node_id":"MDQ6VXNlcjIxMDMxMDY3","organizations_url":"https://api.github.com/users/Codertocat/orgs","received_events_url":"https://api.github.com/users/Codertocat/received_events","repos_url":"https://api.github.com/users/Codertocat/repos","site_admin":false,"starred_url":"https://api.github.com/users/Codertocat/starred{/owner}{/repo}","subscriptions_url":"https://api.github.com/users/Codertocat/subscriptions","type":"User","url":"https://api.github.com/users/Codertocat"},"private":false,"pulls_url":"https://api.github.com/repos/Codertocat/Hello-World/pulls{/number}","pushed_at":"2019-05-15T15:20:57Z","releases_url":"https://api.github.com/repos/Codertocat/Hello-World/releases{/id}","size":0,"ssh_url":"git@github.com:Codertocat/Hello-World.git","stargazers_count":0,"stargazers_url":"https://api.github.com/repos/Codertocat/Hello-World/stargazers","statuses_url":"https://api.github.com/repos/Codertocat/Hello-World/statuses/{sha}","subscribers_url":"https://api.github.com/repos/Codertocat/Hello-World/subscribers","subscription_url":"https://api.github.com/repos/Codertocat/Hello-World/subscription","svn_url":"https://github.com/Codertocat/Hello-World","tags_url":"https://api.github.com/repos/Codertocat/Hello-World/tags","teams_url":"https://api.github.com/repos/Codertocat/Hello-World/teams","trees_url":"https://api.github.com/repos/Codertocat/Hello-World/git/trees{/sha}","updated_at":"2019-05-15T15:21:03Z","url":"https://api.github.com/repos/Codertocat/Hello-World","watchers":0,"watchers_count":0},"sender":{"avatar_url":"https://avatars1.githubusercontent.com/u/21031067?v=4","events_url":"https://api.github.com/users/Codertocat/events{/privacy}","followers_url":"https://api.github.com/users/Codertocat/followers","following_url":"https://api.github.com/users/Codertocat/following{/other_user}","gists_url":"https://api.github.com/users/Codertocat/gists{/gist_id}","gravatar_id":"","html_url":"https://github.com/Codertocat","id":21031067,"login":"Codertocat","node_id":"MDQ6VXNlcjIxMDMxMDY3","organizations_url":"https://api.github.com/users/Codertocat/orgs","received_events_url":"https://api.github.com/users/Codertocat/received_events","repos_url":"https://api.github.com/users/Codertocat/repos","site_admin":false,"starred_url":"https://api.github.com/users/Codertocat/starred{/owner}{/repo}","subscriptions_url":"https://api.github.com/users/Codertocat/subscriptions","type":"User","url":"https://api.github.com/users/Codertocat"}}`
